#include "ingame.h"
#include "title.h"
#include "scene.h"

#include "townengine/game_api.h"

#define STB_PERLIN_IMPLEMENTATION
#include <stb_perlin.h>


static void ingame_tick(struct state *state) {
    struct scene_ingame *scn = (struct scene_ingame *)state->scene;

    world_drawdef(scn->world);
    player_calc(scn->player);

    if (input_is_mouse_captured(&ctx.input)) {
        const float sensitivity = 0.6f; /* TODO: put this in a better place */
        scn->yaw += (float)ctx.input.mouse_relative_position.x * sensitivity;
        scn->pitch -= (float)ctx.input.mouse_relative_position.y * sensitivity;
        scn->pitch = clampf(scn->pitch, -89.0f, 89.0f);

        const float yaw_rad = scn->yaw * (float)DEG2RAD;
        const float pitch_rad = scn->pitch * (float)DEG2RAD;

        scn->cam.target = m_vec_norm(((t_fvec3){
            cosf(yaw_rad) * cosf(pitch_rad),
            sinf(pitch_rad),
            sinf(yaw_rad) * cosf(pitch_rad)
        }));
    }

    const t_fvec3 right = m_vec_norm(m_vec_cross(scn->cam.target, scn->cam.up));
    const float speed = 0.04f; /* TODO: put this in a better place */
    if (input_is_action_pressed(&ctx.input, "player_left"))
        scn->cam.pos = fvec3_sub(scn->cam.pos, m_vec_scale(right, speed));

    if (input_is_action_pressed(&ctx.input, "player_right"))
        scn->cam.pos = fvec3_add(scn->cam.pos, m_vec_scale(right, speed));

    if (input_is_action_pressed(&ctx.input, "player_forward"))
        scn->cam.pos = fvec3_add(scn->cam.pos, m_vec_scale(scn->cam.target, speed));

    if (input_is_action_pressed(&ctx.input, "player_backward"))
        scn->cam.pos = fvec3_sub(scn->cam.pos, m_vec_scale(scn->cam.target, speed));

    if (input_is_action_pressed(&ctx.input, "player_jump"))
        scn->cam.pos.y += speed;

    if (input_is_action_pressed(&ctx.input, "player_run"))
        scn->cam.pos.y -= speed;

    /* toggle mouse capture with end key */
    if (input_is_action_just_pressed(&ctx.input, "mouse_capture_toggle")) {
        input_set_mouse_captured(&ctx.input, !input_is_mouse_captured(&ctx.input));
    }

    m_sprite(m_set(path,  "/assets/player/baron-walk.png"),
             m_set(rect,  ((t_frect){ 256, 256, 48, 48 })),
             m_opt(color, ((t_color){ 255, 255, 255, 255 })),
             m_opt(stretch, false ),
             m_opt(texture_origin, ((t_fvec2){ ctx.tick_count % 48, ctx.tick_count % 48 })));

    m_sprite(m_set(path,  "/assets/light.png"),
             m_set(rect,  ((t_frect){ 48, 64, 64, 64 })),
             m_opt(color, ((t_color){ 255, 0, 0, 255 })));

    m_sprite(m_set(path,  "/assets/light.png"),
             m_set(rect,  ((t_frect){ 64, 64, 64, 64 })),
             m_opt(color, ((t_color){ 0, 255, 0, 255 })));

    m_sprite(m_set(path,  "/assets/light.png"),
             m_set(rect,  ((t_frect){ 80, 64, 64, 64 })),
             m_opt(color, ((t_color){ 0, 0, 255, 255 })));

    m_sprite(m_set(path,     "/assets/player/baron-walk.png"),
             m_set(rect,     ((t_frect){ 32, 32, 64, 64 })),
             m_opt(rotation, (float)M_PI * 2 * (float)(ctx.tick_count % 64) / 64 ));

    m_sprite(m_set(path,     "/assets/player/baron-walk.png"),
             m_set(rect,     ((t_frect){ 128, 32, 128, 64 })),
             m_opt(rotation, (float)M_PI * 2 * (float)(ctx.tick_count % 64) / 64 ));

    set_camera(&scn->cam);

    #define TERRAIN_FREQUENCY 0.1f

    for (int y = 64; y--;) {
        for (int x = 64; x--;) {
            float d0 = stb_perlin_noise3((float)x       * TERRAIN_FREQUENCY, (float)y       * TERRAIN_FREQUENCY, 0, 0, 0, 0) * 20 - 6;
            float d1 = stb_perlin_noise3((float)(x + 1) * TERRAIN_FREQUENCY, (float)y       * TERRAIN_FREQUENCY, 0, 0, 0, 0) * 20 - 6;
            float d2 = stb_perlin_noise3((float)(x + 1) * TERRAIN_FREQUENCY, (float)(y - 1) * TERRAIN_FREQUENCY, 0, 0, 0, 0) * 20 - 6;
            float d3 = stb_perlin_noise3((float)x       * TERRAIN_FREQUENCY, (float)(y - 1) * TERRAIN_FREQUENCY, 0, 0, 0, 0) * 20 - 6;

            unfurl_triangle("/assets/grass.gif",
                            (t_fvec3){ (float)x,     d0, (float)y },
                            (t_fvec3){ (float)x + 1, d1, (float)y },
                            (t_fvec3){ (float)x,     d3, (float)y - 1 },
                            (t_shvec2){ 0, 768 },
                            (t_shvec2){ 1024, 768 },
                            (t_shvec2){ 1024, 0 });

            unfurl_triangle("/assets/grass.gif",
                            (t_fvec3){ (float)x + 1, d1, (float)y },
                            (t_fvec3){ (float)x + 1, d2, (float)y - 1 },
                            (t_fvec3){ (float)x,     d3, (float)y - 1 },
                            (t_shvec2){ 1024, 0 },
                            (t_shvec2){ 0, 0 },
                            (t_shvec2){ 0, 768 });
        }
    }
}


static void ingame_end(struct state *state) {
    struct scene_ingame *scn = (struct scene_ingame *)state->scene;
    player_destroy(scn->player);
    world_destroy(scn->world);
    free(state->scene);
}


struct scene *ingame_scene(struct state *state) {
    (void)state;

    struct scene_ingame *new_scene = ccalloc(1, sizeof *new_scene);
    new_scene->base.tick = ingame_tick;
    new_scene->base.end = ingame_end;

    new_scene->world = world_create();
    new_scene->player = player_create(new_scene->world);

    new_scene->cam = (t_camera){ .pos = { 32, 0, 1 }, .up = { 0, 1, 0 }, .fov = (float)M_PI_2 };

    play_audio_ex("music/mod65.xm", "soundtrack", (t_play_audio_args){
        .repeat = true,
        .volume = 1.0f
    });

    input_set_mouse_captured(&ctx.input, true);

    return (struct scene *)new_scene;
}
